/*-------------------------------------------------------------------
 * nfpt: no-frills parallel transfer
 * a quick-and-dirty pseudo-ECP mode parallel cable transfer utility
 * Copyright (c) 2002,2003 K.H. Man <mkh@pl.jaring.my>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * ecp header file
 * Last revised: 2003-07-15
 *-------------------------------------------------------------------
 * WARNING: no test done for a valid REGbase, it MUST be set first
 * WARNING: no test done for valid arguments, user's responsibility
 * WARNING: rawclock() _might_ have a rare overflow condition
 * WARNING: no guarantees the protocol can work perfectly everywhere
 * WARNING: MAX_PACKET is limited by slices, which is a BYTE!!!
 * WARNING: very wacky handling of CRC and checksum
 *
 * For an example on how to use the calls, please see the test main()
 *-------------------------------------------------------------------
 */
// #define DEBUG

#include <pc.h>
#include <time.h>

/*-------------------------------------------------------------------
 * - custom handshaking for this library (because I was too lazy to
 *   follow the protocol in the Microsoft document)
 * - the idea is that when both sides are ready for a transfer, the
 *   signals are at the recommended ECP start state
 * - furthermore, nInit=0 implies read, nInit=1 implies write
 * - then, we work backwards to devise a signalling convention
 * - non-ECP signal names are used, the old names seem to be more
 *   commonly used than new ECP-type names
 * - there is a glitch in receiving (a 0xFF is always encountered)
 *   I have not sorted that one out yet but worked around it
 *   (which is really not a very wise thing to do)
 *
 *                              (nAck)  (busy)  (PError)  <-- Read
 *                              nStrobe nAutoFd nInit     <-- Write
 * (Tx) BASE        listen          0       1       0
 *      SENDS       getready        0       1       1
 *      WAITS FOR   ack             1       1       0
 *      SETS        start           1       1       1
 *      TRANSMITS   writeBlock      <ecp mode>
 *      WAITS FOR   complete        0       1       0
 *      SETS        listen          0       1       0
 *
 * (Rx) BASE        listen          0       1       0
 *      WAITS FOR   incoming        0       1       1
 *      SENDS       ready           1       1       0
 *      RECEIVES    receiveBlock    <ecp mode>
 *      SETS        listen          0       1       0
 *
 *  (Tx) getready   =>  (Rx) incoming
 *  (Tx) ack        <=  (Rx) ready
 *
 * * when setting registers, nStrobe and nAutoFd are inverted
 * * when reading registers, busy is inverted
 *
 * Pin assignments for handshake signals:
 *  nstrobe(01)  --> nAck(10)
 *  nAutoFd(14)  --> busy(11)
 *  nInit(16)    --> PError(12)
 *-------------------------------------------------------------------
 */

/*-------------------------------------------------------------------
 * Defines method of transfer, for testing, performance optimization
 * PORT_IO_LOOP     basic loop, not the fastest style...
 * PORT_IO_CALL     call a string i/o library function
 * PORT_IO_OPT      optimized code or assembly
 *-------------------------------------------------------------------
 */

#define PORT_IO_LOOP
// #define PORT_IO_CALL
// #define PORT_IO_OPT

/*-------------------------------------------------------------------
 * Test parameters
 *-------------------------------------------------------------------
 */

#define BLK_SIZE    250     /* 4000 bytes if 16 byte FIFO */
#define BLK_COUNT   300     /* number of blocks to send = 1,200,000B */
                            /* this will take a few seconds */

/*-------------------------------------------------------------------
 * LPT register and value definitions, for readability
 *-------------------------------------------------------------------
 */

#define REGdata         (REGbase+0x000)     /* LPT registers, from */
#define REGecpAFifo     (REGbase+0x000)     /* Microsoft ECP doc */
#define REGdsr          (REGbase+0x001)
#define REGdcr          (REGbase+0x002)
#define REGcFifo        (REGbase+0x400)
#define REGecpDFifo     (REGbase+0x400)
#define REGtFifo        (REGbase+0x400)
#define REGcnfgA        (REGbase+0x400)
#define REGcnfgB        (REGbase+0x401)
#define REGecr          (REGbase+0x402)

#define MODEspp         0                   /* Enhanced LPT modes */
#define MODEps2         1                   /* from MS ECP doc */
#define MODEfifo        2
#define MODEecp         3
#define MODEvendor1     4
#define MODEvendor2     5
#define MODEtest        6
#define MODEconfig      7

#define DIRECTIONwrite  0x00                /* ECP direction */
#define DIRECTIONread   0x20

#define STATnBusy       0x80                /* status bits */
#define STATnAck        0x40                /* merged from dcr,ecr */
#define STATPError      0x20
#define STATselect      0x10
#define STATnFault      0x08
#define STATfull        0x02
#define STATempty       0x01

#define SETBselectIn    0x08                /* dcr bits positions */
#define SETBnInit       0x04
#define SETBautoFd      0x02
#define SETBstrobe      0x01

#define MASKdcr         0xF0                /* dcr bit masks */
#define MASKselectIn    0xF7
#define MASKnInit       0xFB
#define MASKautoFd      0xFD
#define MASKstrobe      0xFE

/*-------------------------------------------------------------------
 * ECP port address, IRQ and DMA lookup tables
 *-------------------------------------------------------------------
 */

#define ECP_LIST        3               /* sizes of lists to test */
#define IRQ_LIST        8
#define DMA_LIST        8

#define IRQdefault      7               /* IRQ and DMA defaults */
#define DMAdefault      1

#ifdef ECP_C
int ECPaddrs[] = { 0x378, 0x278, 0x3BC };   /* lookup tables */
int IRQlookup[] = { 0,7,9,10,11,14,15,5 };
int DMAlookup[] = { 0,1,2,3,0,5,6,7 };

/*-------------------------------------------------------------------
 * global variables for the library
 *-------------------------------------------------------------------
 */

int REGbase;    /* LPT base address, LPT port config info */
int irq;        /* IRQ used */
int dma;        /* DMA used */
int pwordsz;    /* pword size */
int FIFO;       /* FIFO size (a slice unit) */

int pktcount = 0;   /* packet counter (packet id sequencing) */
#endif

/*-------------------------------------------------------------------
 * global defines for the library
 *-------------------------------------------------------------------
 */

#define BYTE            unsigned char
#define ulg             unsigned long   /* for crc routine */

#define MAX_PACKET      4000            /* max size of data portion */
#define ECP_TX_TIMEOUT  5		/* ticks to timeout transmitting */
#define ECP_RX_TIMEOUT  5		/* ticks to timeout receiving */
#define ECP_IO_TIMEOUT  3		/* ticks to timeout in/out routines */
#define ECP_RETRIES     3               /* maximum number of retries */

#define CMDnull         0               /* do nothing */
#define CMDhello        1               /* connects to remote pgm */
#define CMDgoodbye      2               /* disconnects from remote */
#define CMDterminate    3               /* reset after TX/RX error */
#define CMDdirectory    4               /* directory path info */
#define CMDfileinfo     5               /* carries file info */
#define CMDfiledata     6               /* carries file data */
#define CMDack          7               /* ACK, successful RX */
#define CMDnak          8               /* NAK, unsuccessful RX */

#define HEADER_ID       0x67            /* packet header ID */

struct HEADER               /* communications packet header, 16 bytes */
{
    BYTE id;                /* identifier, set by ECPcmdPacket */
    BYTE cmd;               /* command */
    BYTE chksum;            /* checksum, set by ECPchecksum */

    BYTE slicesize;         /* slice size (FIFO) (ECPcmdPacket) */
    BYTE slices;            /* slices of additional data */

    BYTE info1;             /* additional info, defaults to 0 */
    BYTE info2;
    BYTE info3;

    unsigned int crc;       /* 32-bit crc, defaults to 0 */
    int count;              /* packet counter (ECPcmdPacket) */
};

#define HEADER_SIZE     sizeof(struct HEADER)

/*-------------------------------------------------------------------
 * error codes
 *-------------------------------------------------------------------
 */

#define ECP_SUCCESS         0x0000      /* general success */
#define ECP_ERR_TIMEOUT     0x0001      /* run-time: time out */
#define ECP_ERR_CHECKSUM    0x0002      /* run-time: header integrity */
#define ECP_ERR_CRC         0x0003      /* run-time: data integrity */
#define ECP_ERR_NOACK       0x0004      /* run-time: handshaking */
#define ECP_ERR_NOCOMPLETE  0x0005      /* run-time: handshaking */
#define ECP_ERR_NOINCOMING  0x0006      /* run-time: handshaking */
#define ECP_ERR_TRANSMIT    0x0007      /* run-time: packet transmit */
#define ECP_ERR_ACKNOWLEDGE 0x0008      /* run-time: packet acknowledge */
#define ECP_ERR_RECEIVE     0x0009      /* run-time: packet receive */
#define ECP_ERR_FILEIO      0x000A      /* run-time: file i/o error */

#define ECP_ERR_NOT_FOUND   0x000F      /* not found (fatal) */
#define ECP_ERR_CONFLICT    0x0010      /* interrupt conflict */
#define ECP_ERR_DIRECTION   0x0020      /* direction bit problem */
#define ECP_ERR_PWORD8      0x0040      /* PWord must be 8 bits */
#define ECP_ERR_FIFO        0x0080      /* must have a FIFO */
#define ECP_ERR_IRQ         0x0100      /* IRQ set by jumper */
#define ECP_ERR_DMA         0x0200      /* DMA set by jumper */

/*-------------------------------------------------------------------
 * support routines
 *-------------------------------------------------------------------
 */

ulg crc(BYTE *buf, int size);           /* crc calculation */

/*-------------------------------------------------------------------
 * primitives
 *-------------------------------------------------------------------
 */

void ECPmode(int MODE);
void ECPdirection(int DIRECTION);

/*-------------------------------------------------------------------
 * master controls
 *-------------------------------------------------------------------
 */

int ECPtest(void);          /* tests and sets the configuration */
void ECPinit(void);         /* begins operation */
void ECPexit(void);         /* ends operation */

void ECPreset(void);        /* resets port after an error */
int ECPstatus(void);        /* reports ecr,dcr status/signal bits */

/*-------------------------------------------------------------------
 * custom handshaking
 *-------------------------------------------------------------------
 */

void ECPlisten(void);       /* ground state for handshake */
void ECPgetready(void);     /* transmit request */

int ECPincoming(void);      /* listening (read) handshake routines */
int ECPack(void);           /* these are immediate (no waiting) */
int ECPcomplete(void);

void ECPready(void);        /* ack (write) handshake routines */
void ECPstart(void);

/*-------------------------------------------------------------------
 * packet header construction
 *-------------------------------------------------------------------
 */

void ECPcrc(struct HEADER *hd, BYTE *buf, int buflen);  /* set crc */
BYTE ECPchecksum(struct HEADER *hd);                /* calc chksum */

void ECPwriteHeader(struct HEADER *hd, BYTE *buf);  /* header convert */
void ECPreadHeader(struct HEADER *hd, BYTE *buf);

void ECPcmdPacket(struct HEADER *hd, int COMMAND);  /* basic header */

void ECPnullPacket(struct HEADER *hd, int size);    /* form commands */
void ECPhelloPacket(struct HEADER *hd);
void ECPgoodbyePacket(struct HEADER *hd);
void ECPterminatePacket(struct HEADER *hd);

/*-------------------------------------------------------------------
 * main communications routines
 *-------------------------------------------------------------------
 */

int ECPsendPacket(struct HEADER *hd, BYTE *buf, int buflen);
int ECPreceivePacket(struct HEADER *hd, BYTE *buf, int *buflen);

/*-------------------------------------------------------------------
 * transmission and reception routines
 *-------------------------------------------------------------------
 */

int ECPwriteBytes(BYTE *buf, int size);     /* byte-by-byte routine */
int ECPreadBytes(BYTE *buf, int size);

int ECPwriteBlock(BYTE *buf, int size);     /* per-FIFO routines */
int ECPreadBlock(BYTE *buf, int size);

/*-------------------------------------------------------------------
 * makes and checks random test data
 *-------------------------------------------------------------------
 */

BYTE block_make(BYTE *buf, int buflen);
int block_compare(BYTE *buf, int buflen, BYTE seed);

/*-------------------------------------------------------------------*/
