/*-------------------------------------------------------------------
 * nfpt: no-frills parallel transfer
 * a quick-and-dirty pseudo-ECP mode parallel cable transfer utility
 * Copyright (c) 2002,2003 K.H. Man <mkh@pl.jaring.my>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * nfpt main file
 * Last revised: 2003-07-15
 *-------------------------------------------------------------------
 * NOTE: locking of disk buffer memory is not actually needed
 *-------------------------------------------------------------------
 * TODO: command to ask for filename and transmit it
 *-------------------------------------------------------------------
 */

#define NFPT_C
#include "nfpt.h"
#include <dpmi.h>

#undef DEBUG_TX_NOTE	/* one line info of packet transmitter */
#undef DEBUG_RX_NOTE	/* one line info of packet receiver */
#undef DEBUG_TX_TRACE	/* detailed trace of packet transmitter */
#undef DEBUG_RX_TRACE	/* detailed trace of packet receiver */
#undef DEBUG_RX_ROUGE	/* allow reporting of rogue packets */

/*-------------------------------------------------------------------
 * debugging support
 *-------------------------------------------------------------------
 */

#ifdef DEBUG_TX_TRACE
#define trace_tx(msg) \
    printf("[tx=%d] result=%d,e1=%d,e2=%d; ", hd->count, result, errorcount1, errorcount2); \
    printf(msg)
#else
#define trace_tx(msg)
#endif

#ifdef DEBUG_RX_TRACE
#define trace_rx(msg) \
    printf("[tx=%d] result=%d,e=%d; ", hd->count, result, errorcount); \
    printf(msg)
#else
#define trace_rx(msg)
#endif

/*-------------------------------------------------------------------
 * NFPT private variables and structures
 *-------------------------------------------------------------------
 */
 
int TX_total_blocks;    /* transmit and receive totals (blocks) */
int RX_total_blocks;
int TX_total_fails;
int RX_total_fails;
int TX_total_bytes;   	/* Tx/Rx totals (bytes of data) */
int RX_total_bytes;

#define IOBUF_SIZE 37268
char iobuffer[IOBUF_SIZE];

/*-------------------------------------------------------------------
 * fatal error so die, die, die...
 *-------------------------------------------------------------------
 */

void disp_fatal(char *msg)
{
    printf("%s", msg);
    exit(-1);
}

/*-------------------------------------------------------------------
 * non-fatal error, continue...
 *-------------------------------------------------------------------
 */

void disp_error(char *msg)
{
    printf("%s", msg);
}

/*-------------------------------------------------------------------
 * implements the classic rotating character to denote progress,
 * optionally printing a message; if slowdown <= 0, clearline & reset
 *-------------------------------------------------------------------
 */

#define RATE_SLOWDOWN 16
#define print_progress_reset print_progress(NULL, 0)

void print_progress(char *msg, int slowdown)
{
    char rotate_char[4] = {'-', '\\', '|', '/'};
    static int rotate = 0;
    int r;

    printf("\r");
    if (slowdown <= 0)		/* reset mode */
    {
	printf("%78c\r", ' ');
	rotate = 0;
	return;
    }
    if (msg)
    {
	printf("%s ", msg);
    }
    r = rotate / slowdown;
    if (r >= 4) { r = 0; }
    printf("%c\r", rotate_char[r]);
    rotate++;
    if (rotate / slowdown >= 4) { rotate = 0; }
}

/*-------------------------------------------------------------------
 * initialize ECP port, test rate at the start of the program
 *-------------------------------------------------------------------
 */

void port_init(void)
{
    int result;
    char msg[80];
    struct HEADER tx;

    _go32_dpmi_lock_data(iobuffer, IOBUF_SIZE); /* lock file buffer memory */

    /* try to detect ECP port configuration */
    result = ECPtest();
    if (result != ECP_SUCCESS)	/* never allow UNCONFIGURED to pass */
    {
	if (result & ECP_ERR_CONFLICT)
	{
	    result &= ~ECP_ERR_CONFLICT;
	    sprintf(msg, "Warning: interrupt conflict reported, non-fatal\n");
	    disp_error(msg);
	}
	if (result & ECP_ERR_IRQ)
	{
	    result &= ~ECP_ERR_IRQ;
	    sprintf(msg, "Warning: IRQ conflict reported, non-fatal\n");
	    disp_error(msg);
	}
	if (result & ECP_ERR_DMA)
	{
	    result &= ~ECP_ERR_DMA;
	    sprintf(msg, "Warning: DMA conflict reported, non-fatal\n");
	    disp_error(msg);
	}
	if (result != ECP_SUCCESS)
	{
	    sprintf(msg, "Fatal config error: code %4X\n", result);
	    disp_fatal(msg);
	}
    }
    app_state = UNCONNECTED;
    ECPinit();
    if (ECPcomplete())
    {
	ECPhelloPacket(&tx);	/* try to hail the other side */
	result = TXdata(&tx, 0, 0);
	if (result != ECP_SUCCESS)
	{
	    sprintf(msg, "Failed to connect to peer: code %4X\n", result);
	    disp_error(msg);
	    return;			/* other side is not ready, ignore */
	}
    }
    else
    {
	sprintf(msg, "No peer detected\n");
	disp_error(msg);
	return;			/* other side is not present, ignore */
    }
    app_state = CONNECTED;
}

/*-------------------------------------------------------------------
 * tests the data transfer rate, assume port is connected
 *-------------------------------------------------------------------
 */

void rate_test(void)
{
    int ticks, result;
    int tx_count = 0;
    int errorcount = 0;
    struct HEADER tx;
    BYTE tx_data[MAX_PACKET];
    char msg[80];

    if (app_state == UNCONNECTED)
    {
	tested_rate = 0;	/* no test if not initialized */
	return;
    }
    ticks = rawclock();
    print_progress_reset;
    
    while ((rawclock() - ticks) < TEST_DURATION)
    {
	ECPcmdPacket(&tx, CMDnull);
	tx.slices = TEST_PKT_SIZE / FIFO;
	tx.info1 = block_make(tx_data, TEST_PKT_SIZE);
	result = TXdata(&tx, tx_data, TEST_PKT_SIZE);

	print_progress(NULL, RATE_SLOWDOWN);
	if (result != ECP_SUCCESS)
	{
	    errorcount++;
	    if (errorcount >= NFPT_RETRIES)
	    {
		sprintf(msg, "Too many errors rate testing: %4X\n", result);
		disp_error(msg);
		tested_rate = 0;		/* test inconclusive */
		if (app_state == READY)		/* bump status down */
		{
		    app_state = CONNECTED;
		}
		return;
	    }
	}
	else
	{
	    tx_count++;
	}
    }
    tested_rate = ((((TEST_PKT_SIZE * tx_count) / 1024) * 182) / 10)
		  / TEST_DURATION;
    app_state = READY;
}

/*-------------------------------------------------------------------
 * perform reliable transmission of a data block
 *-------------------------------------------------------------------
 */

int TXdata(struct HEADER *hd, BYTE *buf, int buflen)
{
    int result;
    int errorcount1 = 0;                    /* TX retry */
    int errorcount2 = 0;                    /* RX retry */
    int timeout;
    int nfpt_timeout = NFPT_TIMEOUT;

    struct HEADER rx;                       /* received packet */
    BYTE rx_data[MAX_PACKET];		    /* rx dummy buffer */
    int rx_size;
    int header_id = hd->count & 0xFF;	    /* truncated ID */

    for (;;)
    {
        /* ------------------------------------------------
         * TX (send data)
         */
        do
        {
            result = ECPsendPacket(hd, buf, /* send data */
                                   buflen);
#ifdef DEBUG_TX_NOTE
	    printf("result=%d, buflen=%d, id=%d, cmd=%d, slicesize=%d, slices=%d, info1=%d, crc=%d\n",
		   result, buflen, hd->id, hd->cmd, hd->slicesize, hd->slices, hd->info1, hd->crc);
#endif
            if (result != ECP_SUCCESS)
            {
		trace_tx("send failed\n");
                errorcount1++;
                TX_total_fails++;
            }
        }
        while ((result != ECP_SUCCESS) &&   /* until successful tx */
               (errorcount1 < NFPT_RETRIES));/* or too many retries */

        if (errorcount1 >= NFPT_RETRIES)    /* too many retries */
        {
	    trace_tx("too many send retries\n");
            return ECP_ERR_TRANSMIT;
        }
        /* ------------------------------------------------
         * RX (acknowledge)
         */
receive_again:
        timeout = rawclock() + nfpt_timeout;/* start clock */
        while ((rawclock() < timeout) &&    /* wait for ack packet */
               !ECPincoming())		    /* or a timeout */
        {
        }
        if (ECPincoming())                  /* reply from target */
        {
            ECPreceivePacket(&rx, rx_data,  /* get reply packet */
			    &rx_size);	    /* dummy rx buffer */

            if (result != ECP_SUCCESS)      /* packet corrupted */
            {                               /* resend */
		trace_tx("received ack bad\n");
                errorcount2++;
            }
            else if (rx.cmd == CMDack)	    /* packet reply ACK */
            {
		if (rx.info1 == header_id)  /* this is a success */
		{			    /* correct packet ID */
		    TX_total_blocks++;
		    TX_total_bytes += buflen;
		    return ECP_SUCCESS;
		}
		else if (rx.info1 > header_id)
		{
		    trace_tx("ack'ed a future id\n");
		    errorcount2++;	   /* can't ack future ID */
		}
		else /* (rx.info1 < header_id) */
		{
		    trace_tx("ack'ed a past id\n");
		    goto receive_again;	   /* drop obsolete acks */
		}
            }
            else                            /* packet reply NAK */
            {                               /* resend */
		trace_tx("received nak reply\n");
                errorcount2++;
            }
        }
        else                                /* timeout handling */
        {                                   /* resend */
	    trace_tx("ack timeout\n");
	    nfpt_timeout += NFPT_TIMEOUT;
            errorcount2++;
        }
        if (errorcount2 >= NFPT_RETRIES)    /* too many retries */
        {
	    trace_tx("too many protocol retries\n");
            return ECP_ERR_ACKNOWLEDGE;
        }
        TX_total_fails++;

    } /* endfor */
}

/*-------------------------------------------------------------------
 * perform reliable reception of a data block
 *-------------------------------------------------------------------
 */

int RXdata(struct HEADER *hd, BYTE *buf, int *buflen)
{
    int result;
    int reply;
    int errorcount = 0;                     /* RX retry */
    int timeout;
    int nfpt_timeout = NFPT_TIMEOUT;
    static int last_count = 0;              /* avoids duplicate blocks */
    struct HEADER tx;                       /* acknowledge packet */

receive_again:
    do
    {
        /* ------------------------------------------------
         * RX (receive data)
         */
        timeout = rawclock() + nfpt_timeout; /* start clock */
        while (!ECPincoming())              /* wait for ack packet */
        {
            if (rawclock() > timeout)       /* process timeout error */
            {
		trace_rx("incoming timeout\n");
		nfpt_timeout += NFPT_TIMEOUT;
                timeout = rawclock() + nfpt_timeout;
                errorcount++;
                RX_total_fails++;
                
                if (errorcount >= NFPT_RETRIES)
                {
		    trace_rx("too many incoming retries\n");
                    return ECP_ERR_RECEIVE;
                }
            }
        }
        result = ECPreceivePacket(hd, buf,  /* receive data */
                                  buflen);
#ifdef DEBUG_RX_NOTE
    	printf("result=%d, buflen=%d, id=%d, cmd=%d, slicesize=%d, slices=%d, info1=%d, crc=%d\n",
	       result, *buflen, hd->id, hd->cmd, hd->slicesize, hd->slices, hd->info1, hd->crc);
#endif
        /* ------------------------------------------------
         * TX (send acknowledge)
         */
        if (result != ECP_SUCCESS)          /* decide on reply */
        {                                   /* ack or nak */
	    trace_rx("receive packet error\n");
            reply = CMDnak;
            errorcount++;
            RX_total_fails++;
        }
        else
        {
            reply = CMDack;
        }
        ECPcmdPacket(&tx, reply);           /* form reply packet */
	tx.info1 = hd->count;		    /* ack the packet ID, truncated!! */

        ECPsendPacket(&tx, 0, 0);           /* send acknowledgement */
                                            /* ignore errors */

        if (hd->count == last_count)        /* same packet still being */
        {                                   /* sent? out of sync... */
	    trace_rx("identical id as last_count\n");
            errorcount = 0;
            goto receive_again;
        }
        if (errorcount >= NFPT_RETRIES)      /* too many retries */
        {
	    trace_rx("too many receive retries\n");
            return ECP_ERR_RECEIVE;
        }
    } while (reply == CMDnak);

    RX_total_blocks++;
    RX_total_bytes += *buflen;
    last_count = tx.info1;                  /* remember packet ID */

    return ECP_SUCCESS;
}

/*-------------------------------------------------------------------
 * perform reliable transmission of file data
 *-------------------------------------------------------------------
 */

int TXfile(struct TAGGEDFILE *tf)
{
    FILE *ifile;
    BYTE tx_data[MAX_PACKET];
    struct HEADER tx;
    int fileinfo_size;
    int block_full, block_rem, block_total, tx_size;
    int result, i;
    char msg[256];
    char *fname = basename(tf->st_name);
    
    /* check fileinfo header packet (tf) */
    if (tf->st_size == 0)
    {
	sprintf(msg, "Transmit error file size zero: %s\n", fname);
	disp_error(msg);
	return ECP_ERR_FILEIO; 		/* nothing to transfer, ignore */
    }
    ifile = fopen(tf->st_name, "rb");	/* open the file for reading */
    if (!ifile)
    {
	sprintf(msg, "Transmit error cannot open, file: %s\n", fname);
	disp_error(msg);
	return ECP_ERR_FILEIO;
    }
    setbuf(ifile, iobuffer);		/* try a locked file buffer */
    
    ECPcmdPacket(&tx, CMDfileinfo);
    block_full = tf->st_size / MAX_PACKET;
    block_rem = tf->st_size - block_full * MAX_PACKET;
    block_total = block_full;
    if (block_rem) { block_total++; }
    /* printf("%d, %d, %d\n", block_total, block_full, block_rem); */

    result = TXdata(&tx, (BYTE *)tf, sizeof(struct TAGGEDFILE));
    if (result != ECP_SUCCESS)
    {
	sprintf(msg, "Transmit header failed, file: %s\n", fname);
	disp_error(msg);
	fclose(ifile);
	return ECP_ERR_TRANSMIT;
    }
    print_progress_reset;
    
    for (i = 0; i < block_total; i++)
    {
	if (block_rem && i == block_full) {
	    tx_size = block_rem;
	}
	else {
	    tx_size = MAX_PACKET;
	}
	result = fread(tx_data, 1, tx_size, ifile);
	if (result != tx_size)
	{
	    sprintf(msg, "Transmit file read failed, file: %s\n", fname);
	    disp_error(msg);
	    fclose(ifile);
	    return ECP_ERR_FILEIO;
	}
	ECPcmdPacket(&tx, CMDfiledata);
	tx.info1 = i;
	/* printf("Tx -- %d - %d - %d\n", i, tx.info1, tx_size); */
	result = TXdata(&tx, tx_data, tx_size);
	
	if (result != ECP_SUCCESS)
	{
	    sprintf(msg, "Transmit data failed, block %d, file: %s\n", i, fname);
	    disp_error(msg);
	    fclose(ifile);
	    return ECP_ERR_TRANSMIT;
	}
	sprintf(msg, "[%s] block %d/%d: ", fname, i+1, block_total);
	print_progress(msg, RATE_SLOWDOWN);
    }
    fclose(ifile);
    return ECP_SUCCESS;
}

/*-------------------------------------------------------------------
 * perform reliable reception of file data
 *-------------------------------------------------------------------
 */

int RXfile(struct HEADER *rx0, struct TAGGEDFILE *tf)
{
    FILE *ofile;
    struct HEADER rx;
    BYTE rx_data[MAX_PACKET];
    int block_full, block_rem, block_total, rx_size, ac_size;
    int result, i, id;
    struct utimbuf newtime;
    char msg[256];
    char *fname = basename(tf->st_name);

    ofile = fopen(tf->st_name, "wb");   /* open the file for writing */
    if (!ofile)
    {
	sprintf(msg, "Receive error cannot open file: %s\n", fname);
	disp_error(msg);
	return ECP_ERR_FILEIO;
    }
    setbuf(ofile, iobuffer);		/* try a locked file buffer */

    block_full = tf->st_size / MAX_PACKET;
    block_rem = tf->st_size - block_full * MAX_PACKET;
    block_total = block_full;
    if (block_rem) { block_total++; }
    /* printf("%d, %d, %d\n", block_total, block_full, block_rem); */
    print_progress_reset;
    
    for (i = 0; i < block_total; i++)
    {
	if (block_rem && i == block_full) {
	    ac_size = block_rem;
	}
	else {
	    ac_size = MAX_PACKET;
	}
	id = i & 0xFF;
retry_receive:
	result = RXdata(&rx, rx_data, &rx_size);
	if (result != ECP_SUCCESS)
	{
	    sprintf(msg, "Receive data failed, block %d, file: %s\n", i, fname);
	    disp_error(msg);
	    fclose(ofile);
	    return ECP_ERR_RECEIVE;
	}
	if (rx.cmd != CMDfiledata)
	{
	    sprintf(msg, "Receive data command error, block %d, file: %s\n", i, fname);
	    disp_error(msg);
	    fclose(ofile);
	    return ECP_ERR_RECEIVE;
	}
	if ((rx.info1 < id) || 		/* awful kludge, drops repeated packets */
	    ((rx.info1 == 255) && (id == 0)))	/* caused an awful bug... */
	{
	    goto retry_receive;
	}
	if (rx_size < ac_size)		/* make sure size is correct */
	{
	    goto retry_receive;
	}
	result = fwrite(rx_data, 1, ac_size, ofile);
	if (result != ac_size)
	{
	    sprintf(msg, "Receive data write error, block %d, file: %s\n", i, fname);
	    disp_error(msg);
	    fclose(ofile);
	    return ECP_ERR_FILEIO;
	}
	sprintf(msg, "[%s] block %d/%d: (%3d,%3d) ", fname, i+1, block_total, id, rx.info1);
	print_progress(msg, RATE_SLOWDOWN);
    }
    fclose(ofile);

    chmod(tf->st_name, (tf->st_mode & S_IRUSR) |    /* permissions */
		       (tf->st_mode & S_IWUSR));
    newtime.actime = tf->st_mtime;		    /* set mod time to st_mtime */
    newtime.modtime = tf->st_mtime;
    utime(tf->st_name, &newtime);
    return ECP_SUCCESS;
}

/*-------------------------------------------------------------------
 * performs reception tasks
 *-------------------------------------------------------------------
 */

void process_incoming(void)
{
    struct HEADER rx;
    BYTE rx_data[MAX_PACKET];
    struct TAGGEDFILE *tf;
    int time_elapsed;
    double per_file_rate;
    int rx_size;
    int result;
    char msg[256];

receive_more:
    result = RXdata(&rx, rx_data, &rx_size);

    if ((result != ECP_SUCCESS) && (rx.cmd != CMDnull))
    {
	/* error, exception is we ignore null packet in error */
	sprintf(msg, "Packet receive error: cmd %d code %4X\n", rx.cmd, result);
	disp_error(msg);
	return;
    }
    switch (rx.cmd)
    {
        /* ------------------------------------------------ */
	case CMDhello:
	    /* hail from peer, we have a connection */
	    if ((app_state == CONNECTED) || (app_state == READY))
	    {
		printf("Peer hails again from remote machine...\n");
	    }
	    else
	    {
		printf("Peer hails from remote machine...\n");
	    }
	    break;
        /* ------------------------------------------------ */
	case CMDgoodbye:
	    /* do nothing, sender might be connecting again */
	    if ((app_state == CONNECTED) || (app_state == READY))
	    {
		printf("Unconnected from peer...\n");
	    }
	    else
	    {
		printf("Duplicate unconnect packet received...\n");
	    }
	    app_state = UNCONNECTED;
	    break;
        /* ------------------------------------------------ */
	case CMDnull:
	    if (app_state != READY)
	    {
		app_state = READY;	/* assume packets are transfer test */
		printf("Transfer rate testing (see transmitter)...\n");
	    }
	    /* mostly ignored, error in null packets also ignored */
#ifdef DEBUG_RX_ROUGE
	    if (rx.slices == TEST_PKT_SIZE / FIFO)	/* report for fun */
	    {
		int errors = block_compare(rx_data, rx_size, rx.info1);
		if (errors > 0)
		{
		    printf("Null packet data error(s)...\n");
		}
	    }
#endif
	    break;
        /* ------------------------------------------------ */
	case CMDfileinfo:
	    /* received file header packet, start file reception */
	    time_elapsed = rawclock();
	    tf = (struct TAGGEDFILE *)rx_data;

	    /* printf("Receiving file %s, size %d...\n", tf->st_name, tf->st_size); */
	    result = RXfile(&rx, tf);

	    if (result == ECP_SUCCESS)		/* this does actual rx */
	    {
		/* report some receive information here */
		time_elapsed = rawclock() - time_elapsed;
		if (time_elapsed == 0) { time_elapsed++; }
		per_file_rate = (tf->st_size / 1024 * 182) / time_elapsed / 10;
	
		printf("Done! [%s] %d bytes, %.1f sec, %d KB/sec\n",
		       tf->st_name, tf->st_size,
		       (double)time_elapsed / 18.2, (int)per_file_rate);
	    }
	    else if (result == ECP_ERR_FILEIO)
	    {
		printf("File I/O Error: failed to receive file %s\n");
	    }
	    else /* assume ECP_ERR_RECEIVE */
	    {
		printf("Receiving Error: failed to receive file %s\n");
	    }
	    break;
        /* ------------------------------------------------ */
	case CMDfiledata:
	    /* rogue packet, print a warning and then ignore */
#ifdef DEBUG_RX_ROUGE
	    printf("Rogue CMDfiledata packet received, ignoring...\n");
#endif
	    break;
        /* ------------------------------------------------ */
	case CMDack:
	case CMDnak:
	    /* protocol leaking acknowledgement packets */
#ifdef DEBUG_RX_ROUGE
	printf("Rogue CMD{ack|nak} packet received, ignoring...\n");
#endif
	    break;
        /* ------------------------------------------------ */
	case CMDterminate:
	    /* transfer termination packet leaked */
#ifdef DEBUG_RX_ROUGE
	printf("Rogue CMDtermination packet received, ignoring...\n");
#endif
	    break;
        /* ------------------------------------------------ */
	default:
#ifdef DEBUG_RX_ROUGE
	    printf("Unrecognized packet received, ignoring...\n");
#endif
	    break;
    }
    if (ECPincoming())
    {
	goto receive_more;
    }
}

/*-------------------------------------------------------------------
 * performs transmission tasks
 *-------------------------------------------------------------------
 */

int process_outgoing(char *fname)
{
    int result, time_elapsed = 0;
    double per_file_rate = 0;
    struct stat fs;
    struct TAGGEDFILE tf;

    if (stat(fname, &fs))		/* cannot detect file */
    {
	printf("File I/O Error: cannot detect file %s\n", fname);
	return 0;
    }
    if (S_ISDIR(fs.st_mode))		/* not a file */
    {
	printf("File I/O Error: %s is not a file\n", fname);
	return 0;
    }
    if (fs.st_size == 0)		/* zero length file */
    {
	printf("File I/O Error: file %s has size of zero\n", fname);
	return 0;
    }
    strcpy(tf.st_name, fname);		/* build a file info struct */
    tf.st_ctime = fs.st_ctime;
    tf.st_mode = fs.st_mode;
    tf.st_mtime = fs.st_mtime;
    tf.st_size = fs.st_size;

    time_elapsed = rawclock();
    result = TXfile(&tf);		/* this does actual tx */
    if (result == ECP_ERR_FILEIO)
    {
	printf("File I/O Error: failed to transmit file %s\n", fname);
	return 0;
    }
    else if (result == ECP_ERR_TRANSMIT)
    {
	printf("Transmitting Error: failed to transmit file %s\n", fname);
	return 0;
    }
    time_elapsed = rawclock() - time_elapsed;
    if (time_elapsed == 0) { time_elapsed++; }
    per_file_rate = (fs.st_size / 1024 * 182) / time_elapsed / 10;

    printf("Done! [%s] %d bytes, %.1f sec, %d KB/sec\n",
	   fname, fs.st_size,
	   (double)time_elapsed / 18.2, (int)per_file_rate);

    return fs.st_size;
}

/*-------------------------------------------------------------------
 * main program loop
 *-------------------------------------------------------------------
 */

int main(int argc, char **argv)
{
    int result, i;
    int time_total = 0, total_size = 0;
    double average_rate = 0;
    struct HEADER tx;
    char msg[256];

    /* -------------------------------------------------------- */
    printf(
    "-------------------------------------------------------------------\n"
    "NFPT: No-Frills Parallel Transfer\n"
    "A quick-and-dirty pseudo-ECP mode parallel cable transfer utility\n"
    "Copyright (c) 2002,2003 K.H. Man <mkh@pl.jaring.my> under the terms\n"
    "of the GNU General Public License Version 2. Compiled <" __DATE__ ">\n"
    "Warning: this is an ALPHA quality prototype!!! Version " NFPT_VER "\n"
    "-------------------------------------------------------------------\n"
    );

    /* -------------------------------------------------------- */
    /* startup parallel port */
    port_init();                    /* initialize and test port, always good */
    printf("ECP port detected...\n");
    printf("Port Address:  0x%03X\t", REGbase);
    printf("DMA Channel:   %2d\n", dma);
    printf("IRQ Number:       %2d\t", irq);
    printf("FIFO Size:     %2d bytes\n", FIFO);
    printf("------------------------------------------------------------------\n");

    /* -------------------------------------------------------- */
    /* data transfer test only if connected to a peer (hello acknowledged) */
    if (app_state == CONNECTED)
    {
	rate_test();
	if (!tested_rate)	/* cannot get a test score */
	{
	    sprintf(msg, "Transmit test failed.\n");
	    disp_error(msg);
	}
	printf("Transfer rate: %d KB/sec\n", tested_rate);
    }
    
    /* -------------------------------------------------------- */
    /* get list of arguments (globbed by DJGPP) */
    if (argc > 1)
    {
	/* make sure connected */
	if (app_state == UNCONNECTED)
	{
	    printf("Cannot transmit, no peer detected.\n");
	}
	else if (app_state == CONNECTED)
	{
	    printf("Cannot transmit, cannot perform test transfer.\n");
	}
	else
	{
	    printf("Transmitting %d files...\n", argc - 1);
	    time_total = rawclock();

	    for (i = 1; i < argc; i++)
	    {
		result = process_outgoing(argv[i]);	/* tx a file */
		if (!result)
		{
		}
		else
		{
		    total_size += result;
		}
	    }
	    time_total = rawclock() - time_total;
	    if (time_total == 0) { time_total++; }
	    average_rate = (total_size / 1024 * 182) / time_total / 10;

	    /* report results */
	    printf("Cumulative results...\n");
	    printf("Total packets sent:    %d\n", TX_total_blocks);
	    printf("Total failed packets:  %d\n", TX_total_fails);
	    printf("Total bytes sent:      %d\n", TX_total_bytes);
	    printf("Total time taken:      %.1f sec\n", (double)time_total / 18.2);
	    printf("Average transfer rate: %d KB/sec\n", (int)average_rate);
	}
    }

    /* -------------------------------------------------------- */
    /* wait for incoming next */
    int done = 0, c;
    if (app_state == UNCONNECTED)
    {
	printf("Waiting for incoming connection...\n");
    }
    else if ((app_state == CONNECTED) || (app_state == READY))
    {
	printf("Waiting for incoming data...\n");
    }
    else
    {	/* should not reach this place */
	printf("Initialization state error, exiting.\n");
	exit(-1);
    }

    /* -------------------------------------------------------- */
    /* instruction line */
    printf("Commands: R -- reset | T -- test | S -- stats | Q -- quit\n");

    while (!done)
    {
        while(!kbhit() && !ECPincoming())	/* polling loop */
        {
        }
	if (ECPincoming())		/* needs to receive */
	{
	    process_incoming();		/* RECEIVE */
	}
	else if (kbhit())
	{
	    c = getkey();
	    switch (c)
	    {
	    /* ------------------------------------------------ */
	    case 'R':			/* unconditional reset */
	    case 'r':
		printf("[Reset Command...]\n");
		ECPreset();
		ECPexit();
		port_init();
		ECPinit();
		if (app_state == UNCONNECTED)
		{
		    printf("Port initialized, no peer detected.\n");
		}
		else
		{   /* CONNECTED */
		    printf("Port initialized, peer acknowledged hello packet.\n");
		}
		break;
	    /* ------------------------------------------------ */
	    case 'T':			/* transfer test */
	    case 't':
		printf("[Transfer Test Command...]\n");
		/* check connection status */
		if (app_state == UNCONNECTED)
		{
		    port_init();	/* reset */
		    if (app_state == UNCONNECTED)
		    {
			printf("Peer not detected, cannot test transfer.\n");
		    }
		}
		if ((app_state == CONNECTED) || (app_state == READY))
		{
		    rate_test();
		    printf("Transfer rate: %d KB/sec\n", tested_rate);
		}
		break;
	    /* ------------------------------------------------ */
	    case 'S':                   	/* statistics */
	    case 's':
		printf(
		    "[Statistics Command...]\n"
		    "TX_total_blocks = %d\tRX_total_blocks = %d\n"
		    "TX_total_fails  = %d\tRX_total_fails  = %d\n"
		    "TX_total_bytes  = %d\tRX_total_bytes  = %d\n",
		    TX_total_blocks, RX_total_blocks,
		    TX_total_fails, RX_total_fails,
		    TX_total_bytes, RX_total_bytes
		);
		break;
	    /* ------------------------------------------------ */
	    case 'Q':                   	/* quit */
	    case 'q':
	    case K_Escape:
		printf("[Quit Command...]\n");
		done = 1;
		break;
	    /* ------------------------------------------------ */
	    default:
		printf("Press R, T, S or Q, dude...\n");
		break;
	    }
	}
    }
    /* -------------------------------------------------------- */
    /* transmit goodbye packet */
    if ((app_state == CONNECTED) || (app_state == READY))
    {
	ECPgoodbyePacket(&tx);	/* send goodbye, ignore errors */
	result = TXdata(&tx, 0, 0);
    }
    ECPexit();
    printf("Bye.\n");
    return 0;
}
/* end of program */
